@@ -3,9 +3,26 @@ import aiohttp
 import logging
 import math
 import asyncio
+import time
+import datetime
+from datetime import timezone
 
 logger = logging.getLogger("core.task")
 
+def _generate_storage_path(powerlink_key: str, task_key: str, app_name: str) -> str:
+        """Generate the S3 storage path using reversed timestamp logic."""
+        max_timestamp = 9999999999999  # Maximum 13-digit timestamp
+        original_timestamp = int(time.time() * 1000)  # Get current timestamp in milliseconds
+        reversed_timestamp = max_timestamp - original_timestamp  # Reverse timestamp
+
+        # Ensure timestamp is interpreted in UTC
+        folder_timestamp = datetime.datetime.fromtimestamp(original_timestamp / 1000, tz=timezone.utc).strftime("%Y%m%d%H%M%S")
+
+        # Construct the storage path
+        storage_path = f"{powerlink_key}/{reversed_timestamp}/{task_key}/{app_name}/{folder_timestamp}"
+        print(f"📂 S3 Storage Path: {storage_path}")  # Debugging output
+        return storage_path
+
 #===================================================================================================
 class Core:
     """Handles the calls to core APIs for task authentication, execution, and lifecycle management."""
@@ -34,6 +51,62 @@ class Core:
         self.CORE_API = CORE_API.rstrip("/")
         self.AGENT_NAME = AGENT_NAME
         self.JWT_TOKEN = None
+        self.RUNTIME_VARIABLES: dict[str, str | None] = {}
+    
+    #================================================================================================
+    # Get Runtime Variable by Key
+    #================================================================================================
+    def get_runtime_variable(self, key: str, default: str | None = None) -> str | None:
+        """
+        Retrieve the value of a runtime variable by key.
+
+        Args:
+            key (str): The key of the variable to fetch.
+            default (str | None): Default value if key is not found.
+
+        Returns:
+            str | None: The variable value or default.
+        """
+        if not self.RUNTIME_VARIABLES or not isinstance(self.RUNTIME_VARIABLES, dict):
+            return default
+        return self.RUNTIME_VARIABLES.get(key, default)
+
+    #================================================================================================
+    # Fetch Runtime Variables from Core
+    #================================================================================================
+    async def fetch_runtime_variables(self):
+        """
+        Fetches runtime variables from Core backend and stores them in self.RUNTIME_VARIABLES.
+        """
+
+        if not self.JWT_TOKEN:
+            raise ValueError("JWT_TOKEN is missing")
+        
+        url = f"{self.CORE_API}/v1/powerlink/{self.AGENT_NAME}/runtime-variables"
+        headers = {
+            "Authorization": f"Bearer {self.JWT_TOKEN}",
+            "Referer": self.CORE_API
+        }
+
+        try:
+            async with aiohttp.ClientSession() as session:
+                async with session.get(url, headers=headers) as response:
+                    if response.status == 200:
+                        result = await response.json()
+                        variables = result.get("variables", [])
+
+                        self.RUNTIME_VARIABLES = {
+                            item["key"]: item["value"] for item in variables
+                        }
+
+                        logger.info(f"✅ Loaded {len(self.RUNTIME_VARIABLES)} runtime variables for slug: {self.AGENT_NAME}")
+                    else:
+                        logger.error(f"❌ Failed to fetch runtime variables: {response.status} - {await response.text()}")
+                        raise RuntimeError(f"Failed to fetch runtime variables: {response.status}")
+
+        except Exception as e:
+            logger.error(f"❌ Error loading runtime variables: {e}")
+            raise
 
     #================================================================================================
     # login to Core and get JWT
@@ -77,7 +150,7 @@ class Core:
                         login_json = await login_response.json()
                         if login_response.status == 200 and "jwt" in login_json:
                             self.JWT_TOKEN = login_json["jwt"]
-                            logger.info(f"✅ [LOGIN] Login succeeded on attempt {attempt}")
+                            logger.info(f"✅ [LOGIN] Login succeeded on attempt {attempt}")                            
                             break
                         else:
                             logger.warning(f"⚠️ [LOGIN] Login failed on attempt {attempt}: {login_response.status} - {login_json}")
@@ -90,7 +163,13 @@ class Core:
             else:
                 raise RuntimeError("[LOGIN] ❌ Authentication failed after maximum retry attempts.")
 
+        if self.JWT_TOKEN:
+            await self.fetch_runtime_variables()
+        else:
+            logger.error("❌ [LOGIN] JWT token not received after login attempts.")
+
         return True
+    
     #================================================================================================
     # Connect to Task in Core
     #================================================================================================
@@ -202,7 +281,7 @@ class Core:
     #================================================================================================
     # Submit Task in Core
     #================================================================================================
-    async def submit_task(self, JOB_COOKIE: dict, task_key: str) -> bool:
+    async def submit_task(self, JOB_COOKIE: dict, task_key: str, output: list[dict] | None = None) -> bool:
         """
         Submits a task in Core by calling the `/submit` endpoint.
 
@@ -229,18 +308,28 @@ class Core:
         try:
             logger.info(f">>> [SUBMIT-TASK] Submit Task: {task_key}")
 
-            url = f"{self.CORE_API}/v1/tasks/{task_key}/submit"
+            url = f"{self.CORE_API}/v1/tasks/{task_key}/submit?noRedirect=true"
             headers = {"Referer": f"{self.CORE_API}"}
             request_body = {
                 "data": {
                     "metadata": f"Submitting Task from {self.AGENT_NAME}"
                 }
             }
+            
+            # Add output only if it's a non-empty list
+            if output and isinstance(output, list):
+                request_body["data"]["output"] = output
 
             async with aiohttp.ClientSession(cookies=JOB_COOKIE) as session:
                 async with session.post(url, headers=headers, json=request_body, allow_redirects=False) as response:
                     response_json = await response.json() if "application/json" in response.headers.get("Content-Type", "") else None
                     
+                    if response.status in [302, 301]:
+                        redirect_url = response.headers.get("Location")
+                        if redirect_url and "done" in redirect_url.lower():
+                            logger.info(f"<<<<< ✅ Job Already submitted! Redirected to: {redirect_url}")
+                            return True
+                    
                     if response.status != 200:
                         logger.error(f"❌ [SUBMIT-TASK]Failed to submit task. HTTP Status: {response.status} - {response_json}")
                         raise RuntimeError(f"❌ [SUBMIT-TASK] Failed to submit task. HTTP Status: {response.status} - {response_json}")
@@ -283,7 +372,7 @@ class Core:
         try:
             logger.info(f">>> Reject Task: {task_key}")
 
-            url = f"{self.CORE_API}/v1/tasks/{task_key}/reject"
+            url = f"{self.CORE_API}/v1/tasks/{task_key}/reject?noRedirect=true"
             headers = {"Referer": f"{self.CORE_API}"}
             request_body = {
                 "data": {
@@ -296,11 +385,17 @@ class Core:
                 async with session.post(url, headers=headers, json=request_body, allow_redirects=False) as response:
                     response_json = await response.json() if "application/json" in response.headers.get("Content-Type", "") else None
                     
+                    if response.status in [302, 301]:
+                        redirect_url = response.headers.get("Location")
+                        if redirect_url and "rejected" in redirect_url.lower():
+                            logger.info(f"<<<<< ✅ Job Already Rejected! Redirected to: {redirect_url}")
+                            return True
+  
                     if response.status != 200:
                         logger.error(f" ❌ [REJECT-TASK] Failed to reject task. HTTP Status: {response.status} - {response_json}")
                         raise RuntimeError(f"❌ [REJECT-TASK] Failed to reject task. HTTP Status: {response.status} - {response_json}")
                     
-                    logger.info(f"<<<<< ✅ Task Rejected!")
+                    logger.info(f"<<<<< ✅ Task Rejected Successfully!")
                     
                     return True
 
@@ -420,7 +515,7 @@ class Core:
     #================================================================================================
     # Create New Job in Core
     #================================================================================================
-    async def new_job(self, workflow_id: int, customer: dict, language: str = "en", valid_until: int = 60) -> dict:
+    async def new_job(self, workflow_id: int, customer: dict, language: str = "en", valid_until: int = 60, metadata: str = None, external_ref_number: str = None, input_data: list = None) -> dict:
         """
         Creates a new job in the Core system using the specified workflow and customer info.
 
@@ -429,6 +524,9 @@ class Core:
             customer (dict): Dictionary containing customer details.
             language (str, optional): Job language code. Defaults to "en".
             valid_until (int, optional): Job expiration in seconds. Defaults to 60.
+            metadata (str, optional): Optional metadata string to include in the job.
+            external_ref_number (str, optional): Optional external reference number.
+            input_data (list, optional): Optional list of input dictionaries.
 
         Returns:
             dict: The created job data returned by the API.
@@ -464,6 +562,18 @@ class Core:
             }
         }
 
+        # 🔧 Optionally add metadata, externalRefNumber, and input to payload
+        if metadata:
+            payload["data"]["metadata"] = metadata
+
+        if external_ref_number:
+            payload["data"]["externalRefNumber"] = external_ref_number
+
+        if input_data:
+            payload["data"]["input"] = input_data
+
+        print(payload)
+
         try:
             async with aiohttp.ClientSession() as session:
                 async with session.post(url, headers=headers, json=payload) as response:
@@ -483,7 +593,7 @@ class Core:
     #================================================================================================
     # Start Job in Core
     #================================================================================================
-    async def start_job(self, job_key: str, access_key: str, metadata: str) -> tuple:
+    async def start_job(self, job_key: str, access_key: str | None = None, metadata: str = "", JOB_COOKIE: dict | None = None) -> tuple:
         """
         Starts a job using jobKey and accessKey, returning job cookies and the response JSON.
 
@@ -495,15 +605,20 @@ class Core:
         Returns:
             tuple: (job_cookie_dict, response_json)
         """
-        if not self.JWT_TOKEN:
-            raise ValueError("[START-JOB] Must login to Core before starting job.")
 
-        if not job_key or not access_key:
-            raise ValueError("[START-JOB] job_key and access_key must be provided.")
+        if not job_key:
+            raise ValueError("[START-JOB] job_key must be provided.")
+        
+        if not access_key and not JOB_COOKIE:
+            raise ValueError("[START-JOB] access_key or JOB_COOKIE must be provided.")
+        
+         # URL depends on cookie presence
+        if JOB_COOKIE:
+            url = f"{self.CORE_API}/v1/jobs/{job_key}/start?noRedirect=true"
+        else:
+            url = f"{self.CORE_API}/v1/jobs/{job_key}/{access_key}/start?noRedirect=true"
 
-        url = f"{self.CORE_API}/v1/jobs/{job_key}/{access_key}/start?noRedirect=true"
         headers = {
-            "Authorization": f"Bearer {self.JWT_TOKEN}",
             "Referer": self.CORE_API
         }
         body = {
@@ -513,10 +628,10 @@ class Core:
         }
 
         try:
-            async with aiohttp.ClientSession() as session:
-                async with session.post(url, headers=headers, json=body) as response:
+            async with aiohttp.ClientSession(cookies=JOB_COOKIE) as session:
+                async with session.post(url, headers=headers, json=body, allow_redirects=False) as response:
                     response_json = await response.json()
-                    if response.status not in [200]:
+                    if response.status != 200:
                         raise RuntimeError(f"❌ [START-JOB] Failed: {response.status} - {response_json}")
 
                     cookies = session.cookie_jar.filter_cookies(self.CORE_API)
@@ -530,3 +645,114 @@ class Core:
         except Exception as e:
             logger.error(f"❌ [START-JOB] Error: {e}")
             raise RuntimeError(f"❌ [START-JOB] Error: {e}")
+
+
+    #================================================================================================
+    # Save File to task
+    #================================================================================================
+    async def save_file(
+        self,
+        JOB_COOKIE: dict,
+        task_key: str,
+        output_key: str,
+        file_name: str,
+        powerlink_key: str,
+        app_name: str,
+        storage_path: str | None = None
+    ):
+        """
+        Uploads a file to a task and appends its key to the task data.
+
+        Args:
+            JOB_COOKIE (dict): Session cookies returned from `connect_task()`.
+            task_key (str): The task key to identify the task.
+            output_key (str): The key under which the file will be stored in task data.
+            file_name (str): The local path of the file to upload.
+            powerlink_key (str): The powerlink key to use for storage path generation.
+            app_name (str): The name of the application uploading the file.
+        
+        Raises:
+            ValueError: If any required parameters are missing or invalid.
+            RuntimeError: If the upload or append operation fails.          
+        """
+        if not JOB_COOKIE or not isinstance(JOB_COOKIE, dict):
+            raise ValueError("❌ [SAVE-FILE] JOB_COOKIE must be a valid non-empty dictionary.")
+
+        if not task_key:
+            raise ValueError("❌ [SAVE-FILE] task_key is required and cannot be empty.")
+        
+        if not output_key:
+            raise ValueError("❌ [SAVE-FILE] output_key is required and cannot be empty.")
+        
+        if not file_name or not os.path.isfile(file_name):
+            raise ValueError(f"❌ [SAVE-FILE] file_name must be a valid file path. Provided: {file_name}")
+        
+        if not powerlink_key:
+            raise ValueError("❌ [SAVE-FILE] powerlink_key is required and cannot be empty.")
+        
+        if not app_name:
+            raise ValueError("❌ [SAVE-FILE] app_name is required and cannot be empty.")
+
+        if not storage_path:
+            storage_path = _generate_storage_path(powerlink_key= powerlink_key, task_key=task_key, app_name=app_name)
+            logger.info(f"📁 No storage_path provided. Generated storage path: {storage_path}")
+        
+        logger.info(f"📁 Saving file to task data: {file_name} at {storage_path}")
+
+        try:
+            # 1. Upload File
+            upload_url = f"{self.CORE_API}/v1/tasks/{task_key}/uploadFile"
+            headers = {"Referer": self.CORE_API}
+
+            async with aiohttp.ClientSession(cookies=JOB_COOKIE) as session:
+                with open(file_name, "rb") as file:
+                    form_data = aiohttp.FormData()
+                    form_data.add_field("file", file, filename=os.path.basename(file_name))
+                    form_data.add_field("path", storage_path)
+                    form_data.add_field("metadata", f"File: [{file_name}] Uploaded by AI Agent: [{self.AGENT_NAME}] into path: [{storage_path}]")
+
+                    async with session.post(upload_url, data=form_data, headers=headers, allow_redirects=False) as upload_resp:
+                        content_type = upload_resp.headers.get("Content-Type", "")
+                        if upload_resp.status in [302, 301]:
+                            redirect_url = upload_resp.headers.get("Location")
+                            logger.error(f"❌ [SAVE-FILE] Session expired or redirected. Redirected to: {redirect_url}")
+                            raise RuntimeError(f"Session expired. Redirected to: {redirect_url}")
+
+                        if "application/json" not in content_type:
+                            text = await upload_resp.text()
+                            logger.error(f"❌ [SAVE-FILE] Unexpected response type: {content_type}. Response: {text}")
+                            raise RuntimeError(f"Unexpected content-type: {content_type}")
+
+                        upload_json = await upload_resp.json()
+                        uploaded_file_key = upload_json.get("data", {}).get("key")
+                        if not uploaded_file_key:
+                            raise RuntimeError("❌ [SAVE-FILE] Upload response missing file key.")
+                        logger.info(f"✅ File uploaded successfully. KEY: {uploaded_file_key}")
+
+            # 2. Append Task Data
+            append_url = f"{self.CORE_API}/v1/tasks/{task_key}/append"
+            append_payload = {
+                "data": {
+                    "metadata": f"App [{app_name}] updated task [{task_key}] with file [{file_name}]",
+                    "output": [
+                        {
+                            "key": output_key,
+                            "value": uploaded_file_key,
+                            "type": "MEDIA",
+                            "description": f"Media file uploaded from app: {app_name}"
+                        }
+                    ]
+                }
+            }
+
+            async with aiohttp.ClientSession(cookies=JOB_COOKIE) as session:
+                async with session.post(append_url, json=append_payload, headers=headers, allow_redirects=False) as append_resp:
+                    if append_resp.status != 200:
+                        text = await append_resp.text()
+                        logger.error(f"❌ [SAVE-FILE] Failed to append task data. Status: {append_resp.status}, Response: {text}")
+                        raise RuntimeError(f"❌ Failed to append task data. Status: {append_resp.status}")
+                    logger.info("✅ Task data appended successfully.")
+
+        except Exception as e:
+            logger.error(f"❌ [SAVE-FILE] Exception: {e}")
+            raise
\ No newline at end of file
